#!/usr/bin/env python

from argparse import ArgumentParser
import cad.dxf as dxf


def generate_dxf():
    """Entry point for generating DXF files.

    """
    parser = ArgumentParser()
    parser.add_argument('--laser-beam-width', type=float,
                        help='Laser beam width in microns',
                        required=True)
    parser.add_argument('--diameter', type=float,
                        help='Circular blob diameter in mm',
                        required=True)
    parser.add_argument('--output-file', type=str,
                        help='DXF file for saving generated grid',
                        required=True)
    args = parser.parse_args()

    laser_beam_width = args.laser_beam_width / 1000.0  # now in mm
    grid_header_str = dxf.grid_header()
    grid_str = dxf.grid(laser_beam_width=laser_beam_width,
                        diameter=args.diameter)
    grid_footer_str = dxf.grid_footer()

    legend_header_str = dxf.legend_header()
    legend_str = dxf.legend(laser_beam_width=laser_beam_width,
                            circle_diameter=args.diameter)
    legend_footer_str = dxf.legend_footer()

    print('Saving grid to file ' + args.output_file)
    with open(args.output_file, 'w') as output_file:
        output_file.write(grid_header_str)
        output_file.write(grid_str)
        output_file.write(grid_footer_str)

    legend_file = dxf.legend_filename(args.output_file)
    print('Saving legend to file ' + legend_file)
    with open(legend_file, 'w') as legend_file:
        legend_file.write(legend_header_str)
        legend_file.write(legend_str)
        legend_file.write(legend_footer_str)



from os.path import join, splitext
from pkg_resources import resource_filename
from math import atan2, cos, sin, sqrt
from numpy import linspace, pi


def __header():
    """Read header in from package data.

    :return:
    """
    header_filename = resource_filename(
        'cad', join('data/dxf', 'header.dxf'))
    with open(header_filename, 'r') as header_file:
        return header_file.read()


def __footer():
    """Read footer in from package data.

    :return:
    """
    footer_filename = resource_filename(
        'cad', join(join('data', 'dxf'), 'footer.dxf'))
    with open(footer_filename, 'r') as footer_file:
        return footer_file.read()


def __polyline():
    """Read polyline in from package data.

    :return:
    """
    polyline_filename = resource_filename(
        'cad', join(join('data', 'dxf'), 'polyline.dxf'))
    with open(polyline_filename, 'r') as polyline_file:
        return polyline_file.read()


def __seqend():
    """Read seqend in from package data.

    :return:
    """
    seqend_filename = resource_filename(
        'cad', join(join('data', 'dxf'), 'seqend.dxf'))
    with open(seqend_filename) as seqend_file:
        return seqend_file.read()


def __format_float(value):
    """Format float in scientific notation, 6 decimal places.

    :param value:
    :return:
    """
    return '{:.6e}'.format(float(value))


def grid_header():
    """Generate grid header text for DXF file.

    :return:
    """
    return __header()


def grid(laser_beam_width, diameter):
    """Generate DXF text describing asymmetrical circles grid.

    :param laser_beam_width:
    :param diameter:
    :return:
    """
    width = 3
    height = 11

    # now we shift paradigm: width becomes rows, height cols
    num_rows = 2 * width
    num_cols_even = height // 2
    num_cols_odd = num_cols_even + 1

    # first and last line coordinates
    x0 = laser_beam_width / 2.0
    y0 = 0.0
    xn = diameter / 2.0

    # number of laser application trajectories (i.e. lines)
    num_lines = int(round(diameter / (2.0 * laser_beam_width)) + 1)

    # step size for lines
    dx = (xn - x0) / (num_lines - 1)

    grid_str = ''
    for row in range(num_rows):
        if row % 2 == 0:  # even row
            num_cols = num_cols_even
            offset = diameter
        else:  # odd row
            num_cols = num_cols_odd
            offset = 0

        for col in range(num_cols):
            for line in range(num_lines):
                grid_str += 'CIRCLE' + '\n'
                for value in [8, 0, 62, 7, 10]:
                    grid_str += str(value) + '\n'
                grid_str += __format_float(
                    x0 + col * (2 * diameter) + offset) + '\n'
                grid_str += str(20) + '\n'
                grid_str += __format_float(
                    y0 + (row + 1) * (1.5 * diameter)) + '\n'
                for value in [30, 0, 40]:
                    grid_str += str(value) + '\n'
                grid_str += __format_float(x0 + dx * line) + '\n'
                grid_str += str(0) + '\n'

    return grid_str


def grid_footer():
    """Generate grid footer text for DXF file.

    :return:
    """
    return __footer()


def legend_filename(grid_filename):
    """Generate legend filename corresponding to `grid_filename`.

    :param grid_filename:
    :return:
    """
    split_buffer = splitext(grid_filename)
    return split_buffer[0] + '-legend' + split_buffer[1]


def legend_header():
    """Generate legend text for DXF file.

    :return:
    """
    return __header()


def legend(laser_beam_width, circle_diameter):
    """Generate DXF text describing legend.

    :param laser_beam_width:
    :param circle_diameter: diameter of circles in
    spanned circles grid, used for scaling legend
    :return:
    """

    # Coordinates of the two vertices of ellipse major axis
    x1 = circle_diameter * -1.26978
    x2 = circle_diameter * 15.6187
    y1 = circle_diameter * 1.68651
    y2 = circle_diameter * 13.8027
    tilt_angle = atan2(y2 - y1, x2 - x1)

    # Centre of ellipse
    xc = (x1 + x2) / 2
    yc = (y1 + y2) / 2

    # Short radius
    b = circle_diameter * 7.84016

    # Total number of passes required
    total_line_width = circle_diameter / 2
    num_lines = int(round(total_line_width / laser_beam_width) + 1)
    mutual_line_distance = total_line_width / (num_lines - 1)

    polyline = __polyline()
    legend_str = ''
    for line in range(num_lines):
        legend_str += polyline

        line_offset = line * mutual_line_distance
        line_offset_x = line_offset * cos(tilt_angle)
        line_offset_y = line_offset * sin(tilt_angle)
        new_x1 = x1 - line_offset_x
        new_x2 = x2 + line_offset_x
        new_y1 = y1 - line_offset_y
        new_y2 = y2 + line_offset_y

        new_a = 0.5 * sqrt((new_x2 - new_x1) ** 2 +
                           (new_y2 - new_y1) ** 2)
        new_b = b + line_offset

        angles = linspace(start=0, stop=2 * pi, num=100)
        for angle in angles:
            pre_tilt_x = new_a * cos(angle)
            pre_tilt_y = new_b * sin(angle)
            x = xc +\
                pre_tilt_x * cos(tilt_angle) -\
                pre_tilt_y * sin(tilt_angle)
            y = yc +\
                pre_tilt_x * sin(tilt_angle) +\
                pre_tilt_y * cos(tilt_angle)

            legend_str += 'VERTEX' + '\n'
            for value in [8, 0, 10]:
                legend_str += str(value) + '\n'
            legend_str += __format_float(x) + '\n'
            legend_str += str(20) + '\n'
            legend_str += __format_float(y) + '\n'
            legend_str += str(0) + '\n'

        legend_str += __seqend()

    return legend_str


def legend_footer():
    """Generate legend footer text for DXF file.

    :return:
    """
    return __footer()



#!/usr/bin/env python

try:
    from cv2 import VideoCapture, imshow, waitKey,\
        putText, FONT_HERSHEY_PLAIN, drawChessboardCorners
except ImportError as e:
    print('OpenCV does not seem to be installed on your system.')
    print('See http://opencv.org for how to install it.')
    print('The detailed error message was:')
    print(str(e))
    quit()

from argparse import ArgumentParser
from numpy import zeros, uint8
import pkg_resources
from os.path import join
import endocal.calibration as calibration

KEY_QUIT = 27
KEY_TOGGLE_ACQUISITION = ord(calibration.State.KEYS[calibration.State.ACQUIRING])
KEY_ABORT_ACQUISITION = ord(calibration.State.KEYS[calibration.State.CORRECTING])


def __frame_size(video_source_desc):
    cap = VideoCapture(video_source_desc)
    ret, tmp_image = cap.read()
    cap.release()
    if not ret:
        raise RuntimeError('Could not read ' + str(video_source_desc))
    return [0, 0, tmp_image.shape[1], tmp_image.shape[0]]


def __run(video_source_desc, roi, pattern_specs, calibration_file,
          output_folder, max_num_frames = float('inf')):
    full = __frame_size(video_source_desc)
    if roi is None:
        roi = full

    pattern_specs = tuple(pattern_specs)
    calibrator = calibration.Calibrator(pattern_specs=pattern_specs,
                                        file_path=calibration_file,
                                        # i.e. None handled internally:
                                        roi=roi,
                                        full=full)
    state = calibration.State(calibrator)
    source = VideoCapture(video_source_desc)
    file_io = calibration.FileIO(output_folder)

    if not source.isOpened():
        raise RuntimeError('Could not open ' + str(video_source_desc))

    frame = zeros((roi[3], 2 * roi[2], 3), dtype=uint8)
    num_frames = 0
    while True:
        ret, image = source.read()
        if not ret:
            source = VideoCapture(video_source_desc)
            continue
        image = image[roi[1]:roi[1]+roi[3],
                      roi[0]:roi[0]+roi[2]]
        frame[0:roi[3], roi[2]:2*roi[2]] = zeros(image.shape)
        if state.is_correcting():
            ret, corrected_image = calibrator.correct(image)
            if ret:
                frame[0:roi[3], roi[2]:2*roi[2]] = corrected_image
        elif state.is_acquiring():
            if num_frames < max_num_frames:
                ret, blobs = calibrator.append(image, file_io.next_image())
                if ret:
                    num_frames += 1
                    drawChessboardCorners(image, calibrator.pattern_dims, blobs, ret)
        elif state.is_calibrating():
            if calibrator.done():
                state.correcting()

        frame[0:roi[3], 0:roi[2]] = image
        putText(frame, state.what(), (30, 30), FONT_HERSHEY_PLAIN, 2, (0, 0, 255))
        imshow('Optical distortion calibration', frame)

        # user input
        key = waitKey(50)
        if (0xFF & key == KEY_QUIT) or image.size == 0:
            break
        elif 0xFF & key == KEY_TOGGLE_ACQUISITION:
            if state.is_correcting():
                file_io.new_session()
                source.release()
                source = VideoCapture(video_source_desc)
                calibrator.reset()
                state.acquiring()
            elif state.is_acquiring():
                num_frames = 0
                if calibrator.can():
                    calibrator.start(file_io.calibration())
                    state.calibrating()
                else:
                    calibrator.reset()
                    state.correcting()
        elif 0xFF & key == KEY_ABORT_ACQUISITION:
            if state.is_acquiring():
                num_frames = 0
                calibrator.reset()
                state.correcting()

    source.release()


def main():
    # parse arguments
    parser = ArgumentParser()
    parser.add_argument('--input', type=str,
                        help='Video file, video folder or device id (e.g. 1 for /dev/video1)',
                        required=True)
    parser.add_argument('--calibration-file', type=str,
                        help='YAML file with calibration parameters',
                        required=False)
    parser.add_argument('--output-folder', type=str,
                        help='Where to log results',
                        required=True)
    parser.add_argument('--roi', nargs=4, type=int,
                        help='Sub-frame specs: <x> <y> <width> <height>',
                        required=False)
    parser.add_argument('--pattern-specs', nargs=4, type=float,
                        help='Calibration pattern dimensions: <rows> <cols> '
                             '<row_spacing> <col_spacing> (rows a.k.a. width, '
                             'cols a.k.a. height)',
                        required=True)
    args = parser.parse_args()

    # do work
    try:
        video_source_desc = int(args.input)
    except ValueError:
        video_source_desc = args.input

    __run(video_source_desc=video_source_desc, roi=args.roi,
          pattern_specs=args.pattern_specs, calibration_file=args.calibration_file,
          output_folder=args.output_folder)


def test():
    dataset_desc = 'sample_001'
    file_wildcard = 'frame_%03d.jpg'
    data_dir = pkg_resources.resource_filename('endocal', join('data', dataset_desc))
    __run(video_source_desc=join(data_dir, file_wildcard),
          max_num_frames=10,
          roi=None,
          pattern_specs=[3, 11, 3, 1],
          calibration_file=None,
          output_folder='./tmp-' + dataset_desc)



from cv2 import __version__, findCirclesGrid,\
    CALIB_CB_ASYMMETRIC_GRID, calibrateCamera, undistort,\
    imwrite, CALIB_FIX_K4, CALIB_FIX_K5, CALIB_ZERO_TANGENT_DIST
from yaml import dump
import numpy as np
from threading import Thread
from time import strftime
from os.path import join, isdir
from os import makedirs
from random import choice
from string import ascii_uppercase
opencv_version = -1
if __version__.startswith('2'):
    opencv_version = 2
    from cv2 import SimpleBlobDetector
elif __version__.startswith('3'):
    opencv_version = 3
    from cv2 import SimpleBlobDetector_create
else:
    raise RuntimeError('OpenCV version not supported')


class State:

    """Keeps app state, providing query functions.

    See member functions for available states.

    """
    CORRECTING = 1
    CALIBRATING = 2 * CORRECTING
    ACQUIRING = 3 * CORRECTING
    KEYS = {ACQUIRING: 's', CALIBRATING: 's',
            CORRECTING: 'x'}

    def __init__(self, calibrator):
        """Initialise current state to recording.

        :param calibrator: `Calibrator` object for
        use in reporting, e.g. nr. of frames added
        """
        self.current = State.CORRECTING
        self.calibrator = calibrator

    def is_correcting(self):
        """

        :return: ``True`` if app currently online,
        i.e. correcting optical distortions
        """
        return self.current == State.CORRECTING

    def correcting(self):
        """Set current state to correcting.

        """
        self.current = State.CORRECTING

    def is_calibrating(self):
        """

        :return: ``True`` if app currently
        performing calibration
        """
        return self.current == State.CALIBRATING

    def calibrating(self):
        """Set current state to calibrating.

        """
        self.current = State.CALIBRATING

    def is_acquiring(self):
        """

        :return: ``True`` if app currently
        acquiring data for calibration
        """
        return self.current == State.ACQUIRING

    def acquiring(self):
        """Set current state to acquiring.

        """
        self.current = State.ACQUIRING

    def what(self):
        """

        :return: A string representing current
        state
        """
        if self.is_acquiring():
            msg = 'Acquired '
            if self.calibrator.grid_candidates:
                num_frames = len(self.calibrator.grid_candidates)
            else:
                num_frames = 0

            msg += str(num_frames)
            msg += ' frames '
            msg += '(min: '
            msg += str(Calibrator.MIN_FRAME_COUNT)
            msg += '), '
            if num_frames >= Calibrator.MIN_FRAME_COUNT:
                msg += State.KEYS[State.CALIBRATING]
                msg += ': calibrate, '
            msg += State.KEYS[State.CORRECTING]
            msg += ': abort acquisition'
        elif self.is_calibrating():
            msg = 'Performing calibration'
        elif self.is_correcting():
            msg = 'Correcting for optical distortions, '
            msg += State.KEYS[State.CALIBRATING]
            msg += ': start acquisition'
        msg += ', ESC: quit app'
        return msg


class Calibrator:

    """Keeps all data structures pertaining to calibration.

    """

    def __init__(self, pattern_specs, full, file_path=None, roi=None):
        """

        :param pattern_specs: Specs of used calibration
        pattern: ``width, height, row-spacing, col-spacing``
        :param full: Full frame specs: ``x, y, width, height``
        :param file_path: YAML file with calibration params
        :param roi: Sub frame specs: ``x, y, width, height``
        """
        self.pattern_dims = tuple(int(val) for val in pattern_specs[:2])
        self.pattern_spacing = pattern_specs[2:]
        if opencv_version == 2:
            self.detector = SimpleBlobDetector()
        elif opencv_version == 3:
            self.detector = SimpleBlobDetector_create()
        d_y = self.pattern_spacing[0]  # 3.0
        d_x = self.pattern_spacing[1]  # 1.0
        width = int(pattern_specs[0])
        height = int(pattern_specs[1])
        self.grid = np.zeros((width * height, 3), np.float32)
        self.grid[:, :2] = np.mgrid[0:width, 0:height].T.reshape(-1, 2)
        self.grid[:, 0] *= d_y
        self.grid[:, 1] *= d_x
        for start in range(width, width*height, 2*width):
            self.grid[start:start + width, 0] += d_y / 2
        self.grids = None
        self.grid_candidates = None
        self.image_size = None
        self.camera_matrix = None
        self.dist_coeffs = None
        Calibrator.MIN_FRAME_COUNT = 10
        self.calibration_thread = None
        self.roi = full
        if roi is not None:
            self.roi = roi
        self.full = full
        self.reset()
        # TODO file_path
        pass

    def reset(self):
        """Reset all data saved so far.

        """
        self.grids = []
        self.grid_candidates = []
        self.image_size = []
        self.camera_matrix = None  # TODO identity
        self.dist_coeffs = None  # TODO identity (??)
        # TODO
        pass

    def can(self):
        """

        :return: ``True`` if calibration can
        be performed.
        """
        return len(self.grid_candidates) >= Calibrator.MIN_FRAME_COUNT

    def correct(self, image):
        """Correct passed image.

        :param image:
        :return: ``retval`` flag and corrected image
        """
        ret = False
        corrected_image = None
        if self.camera_matrix is not None and \
           self.dist_coeffs is not None:
            corrected_image = undistort(image,
                                        self.camera_matrix,
                                        self.dist_coeffs)
            ret = True
        # TODO - if using identity?
        return ret, corrected_image

    def __calibrate(self, file_path):
        """Perform calibration, to be called as a separate
        thread.

        :param file_path: Where to save resulting calibration
        """
        ret, camera_matrix, dist_coeffs, rvecs, tvecs = \
            calibrateCamera(objectPoints=self.grids,
                            imagePoints=self.grid_candidates,
                            imageSize=self.image_size,
                            cameraMatrix=None,
                            distCoeffs=None,
                            flags=CALIB_FIX_K4 |
                            CALIB_FIX_K5 |
                            CALIB_ZERO_TANGENT_DIST)
        if ret:
            self.camera_matrix = camera_matrix
            self.dist_coeffs = dist_coeffs
            if file_path is not None:
                calibration_file = open(file_path, 'w')
                calibration = dict(camera_matrix=str(self.camera_matrix),
                                   dist_coeffs=str(self.dist_coeffs),
                                   roi=str(self.roi),
                                   full=str(self.full))
                calibration_file.write(dump(calibration,
                                            default_flow_style=False))
                calibration_file.close()

    def start(self, file_path):
        """Start calibration thread.

        :param file_path: Where to save calibration upon
        success
        :return: ``True`` if thread started successfully.
        """
        if self.calibration_thread is None:
            self.calibration_thread = Thread(target=self.__calibrate,
                                             kwargs={'file_path': file_path})
            try:
                self.calibration_thread.start()
            except RuntimeError:
                self.calibration_thread = None
                return False
            else:
                return True
        return True

    def done(self):
        """

        :return: Whether calibration thread has finished.
        """
        if self.calibration_thread is None:
            return True

        if not self.calibration_thread.is_alive():
            self.calibration_thread = None
            return True

        return False

    def append(self, image, file_path=None):
        """Append image as one of frames to be used in
        calibration.

        :param image:
        :param file_path: If passed, `image` is saved
        here
        :return: ``True`` if `image` usable, ``False``
        otherwise, in addition to detected circle grid
        of blobs
        """
        blobs = None
        ret, blobs = findCirclesGrid(image, self.pattern_dims,
                                     blobs, CALIB_CB_ASYMMETRIC_GRID)
        if ret:
            self.image_size = image.shape[:2]  # TODO what about the -1 I saw in sample code?
            self.grid_candidates.append(blobs)
            self.grids.append(self.grid)

        if file_path is not None:
            imwrite(file_path, image)

        return ret, blobs


class FileIO:

    """Manages file IO parameters, like session folders.

    """

    def __init__(self, output_folder):
        """

        :param output_folder: Creates session folders
        inside this one
        """
        self.output_folder = output_folder
        self.session_folder = None
        self.current_image = 1

    def new_session(self):
        """Create a new session folder to save current
        calibration data.

        """
        unique_suffix = ''.join(choice(ascii_uppercase) for _ in range(5))
        session_folder = join(self.output_folder,
                              strftime('%Y-%m-%d-%H-%M-%S') +
                              '-' + unique_suffix)
        makedirs(session_folder)
        if isdir(session_folder):
            self.session_folder = session_folder
        self.current_image = 1

    def calibration(self):
        """

        :return: File path where current calibration is
        to be saved
        """
        if self.session_folder is None:
            return None

        return join(self.session_folder, 'calibration.yml')

    def next_image(self):
        """Increment image count and get file path for
        saving next image.

        :return: File path where an image appended to
        calibration data is to be saved
        """
        if self.session_folder is None:
            return None

        file_path = join(self.session_folder,
                         'frame_' +
                         str(self.current_image).zfill(3) +
                         '.jpg')
        self.current_image += 1
        return file_path



"""A compact GUI application for optical distortion calibration of endoscopes.

See:
https://github.com/gift-surg/endocal
"""

from setuptools import setup
# To use a consistent encoding
from codecs import open
from os import path

doc_dir = path.abspath(path.join(path.dirname(__file__), 'doc'))

# Get the summary
with open(path.join(doc_dir, 'summary.txt'), encoding='utf-8') as f:
    summary = f.read()

# Get the long description
with open(path.join(doc_dir, 'description.txt'), encoding='utf-8') as f:
    description = f.read()
    long_description = '%s\n\n%s' % (summary, description)

setup(
    name='endocal',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='16.08.08rc5',

    description=summary,
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/gift-surg/endocal',

    # Author details
    author='Dzhoshkun Ismail Shakir',
    author_email='d.shakir@ucl.ac.uk',

    # Choose your license
    license='BSD-3-Clause',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Video :: Capture',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python',

        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],

    # What does your project relate to?
    keywords='optical distortion calibration, endoscope, endoscopy, medical imaging,'
             'image processing, biomedical engineering, medical physics,'
             'image-guided interventions',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=['endocal', 'cad'],

    # As recommended in
    # https://docs.python.org/2/distutils/setupscript.html#installing-package-data
    package_dir={'endocal': 'endocal', 'cad': 'cad'},

    py_modules=['endocal.calibration', 'cad.dxf'],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['PyYAML', 'numpy'],

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={'endocal': ['data/sample_001/*'],
                  'cad': ['data/dxf/header.dxf', 'data/dxf/footer.dxf',
                          'data/dxf/polyline.dxf', 'data/dxf/seqend.dxf']},

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'endocal=endocal:main',
            'endocal-test=endocal:test',
            'dxf=cad:generate_dxf'
        ],
    },
)
